"""
Module which generates, compiles and loads Python C-extension which is wrapper around Verilator simulation.
Inputs are Verilog files and result is Python class of simulator which can be instantiated and simulated.
"""